// Test utility to run a Windows program within a harness which has installed
//  - an exception handler
//  - a replacement for Windows API call CreateFile() (used for fopen() etc.)
//
// Each time CreateFile() is called it causes spurious access violations
// which should be ignored/passed by the tested application and handled
// by the exception handlers in the detour harness.
//
// This is intended to provide a regression test for the infamous wxvault.dll
// Embassy Trust Security suite issue seen on certain Dell laptops.
//
// Based on samples bundled with "Detours Express 2.1"

#include <stdio.h>
#include <windows.h>
#include "detours.h"
#include "firstexc.h"

#define DLLNAME "exceptionTest.dll"

VOID NullExport() { }


//
// Data accessed to cause access violations
//
static LPVOID   pData1 = NULL;  // Exceptions for accesses here are caught and handled by first chance exception filter
static LPVOID   pData2 = NULL;  // Exceptions for accesses here are caught by structured exception handler (__try/__catch)
static DWORD    perm1 = 0;
static DWORD    perm2 = 0;

static DWORD    exceptionsTriggered = 0;
static DWORD    exceptionsHandled = 0;


//
// Detour for first chance exceptions
//
LONG WINAPI MyVirtualFaultFilter(PEXCEPTION_POINTERS pException)
{
    PEXCEPTION_RECORD pExceptRec = pException->ExceptionRecord;

    if (pExceptRec->ExceptionCode == EXCEPTION_ACCESS_VIOLATION) {
        if (pExceptRec->NumberParameters >= 2 &&
            pExceptRec->ExceptionInformation[1] >= (ULONG)pData1 &&
            pExceptRec->ExceptionInformation[1] <= (ULONG)pData1 + sizeof(ULONG)) {

            VirtualProtect(pData1, sizeof(ULONG), PAGE_READWRITE, &perm1);
            printf(DLLNAME ": First chance handler caught memory access exception.\n");
            exceptionsHandled++;
            return EXCEPTION_CONTINUE_EXECUTION;
        }
    }
    return EXCEPTION_CONTINUE_SEARCH;
}


//
// Exercise first chance exception handling
//
void causeException1(void) {
    // Cause access violation on data area 1. First chance handler gives read/write permission.
    printf(DLLNAME ": Cause exception on data area 1.\n");
    exceptionsTriggered++;
    *(PULONG) pData1 = 1; 

    // Reset page to read-only
    VirtualProtect(pData1, sizeof(ULONG), PAGE_READONLY, &perm1);
}

//
// Exercise structured exception handling
//
void causeException2(void) {
    printf(DLLNAME ": Cause exception on data area 2.\n");
    exceptionsTriggered++;
    BOOL handled = FALSE;
    __try {
        *(PULONG) pData2 = 1; 
    } __except(EXCEPTION_EXECUTE_HANDLER) {
        if ( GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION )
            handled = TRUE;
        // should check addresses here too
    }
    if ( handled ) {
        printf(DLLNAME ": __except() caught memory access exception.\n");
        exceptionsHandled++;
    }
}


//
// Detours for CreateFile API calls (misleadingly also used for fopen(.., "r"))
//
HANDLE (__stdcall * Real_CreateFileA)(LPCSTR a0, DWORD a1, DWORD a2, LPSECURITY_ATTRIBUTES a3, DWORD a4, DWORD a5, HANDLE a6) = CreateFileA;
HANDLE (__stdcall * Real_CreateFileW)(LPCWSTR a0,DWORD a1, DWORD a2, LPSECURITY_ATTRIBUTES a3, DWORD a4, DWORD a5, HANDLE a6) = CreateFileW;

HANDLE __stdcall Mine_CreateFileA(LPCSTR a0, DWORD a1, DWORD a2, LPSECURITY_ATTRIBUTES a3, DWORD a4, DWORD a5, HANDLE a6) {
    printf(DLLNAME ": CreateFileA(%hs,...)\n", a0);

    // Cause access violations
    causeException1();
    causeException2();

    // Do the real operation
    return Real_CreateFileA(a0, a1, a2, a3, a4, a5, a6);
}

HANDLE __stdcall Mine_CreateFileW(LPCWSTR a0, DWORD a1, DWORD a2, LPSECURITY_ATTRIBUTES a3, DWORD a4, DWORD a5, HANDLE a6) {
    printf(DLLNAME ": CreateFileW(%ls,...)\n", a0);

    // Cause access violations
    causeException1();
    causeException2();

    // Do the real operation
    return Real_CreateFileW(a0, a1, a2, a3, a4, a5, a6);
}


//
// Install CreateFile and exception detours
//
BOOL WINAPI DllMain(HINSTANCE hinst, DWORD dwReason, LPVOID reserved)
{
    LONG error;
    (void)hinst;
    (void)reserved;

    if (dwReason == DLL_PROCESS_ATTACH) {
        printf(DLLNAME ": Starting.\n");
        fflush(stdout);

        // Allocate and protect memory accessed to cause internal access violations
        pData1 = VirtualAlloc(NULL, sizeof(ULONG), MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE);
        if (pData1 == NULL) {
            printf(DLLNAME ": VirtualAlloc failed: %d\n", GetLastError());
            return FALSE;
        }
        *(PULONG)pData1 = 0;
        VirtualProtect(pData1, sizeof(ULONG), PAGE_READONLY, &perm1);

        pData2 = VirtualAlloc(NULL, sizeof(ULONG), MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE);
        if (pData2 == NULL) {
            printf(DLLNAME ": VirtualAlloc failed: %d\n", GetLastError());
            return FALSE;
        }
        *(PULONG)pData2 = 0;
        VirtualProtect(pData2, sizeof(ULONG), PAGE_READONLY, &perm2);

        // Install CreateFile detours
        DetourRestoreAfterWith();

        DetourTransactionBegin();
        DetourUpdateThread(GetCurrentThread());

        DetourAttach(&(PVOID&)Real_CreateFileA, Mine_CreateFileA);
        DetourAttach(&(PVOID&)Real_CreateFileW, Mine_CreateFileW);
        error = DetourTransactionCommit();

        if (error != NO_ERROR) {
            printf(DLLNAME ": Error detouring CreateFile[W|A](): %d\n", error);
            return FALSE;
        }

        printf(DLLNAME ": Detoured CreateFile[W|A]().\n");
        DetourFirstChanceExceptionFilter(MyVirtualFaultFilter);
        printf(DLLNAME ": Detoured first chance exception filter.\n");

    } else if (dwReason == DLL_PROCESS_DETACH) {
        DetourTransactionBegin();
        DetourUpdateThread(GetCurrentThread());

        DetourDetach(&(PVOID&)Real_CreateFileA, Mine_CreateFileA);
        DetourDetach(&(PVOID&)Real_CreateFileW, Mine_CreateFileW);

        error = DetourTransactionCommit();

        printf(DLLNAME ": Removed CreateFile[W|A] detour (result=%d).\n", error);
        if ( exceptionsTriggered == exceptionsHandled ) {
            if ( exceptionsTriggered == 0 ) {
                printf(DLLNAME ": TEST ERROR: No exceptions triggered.\n");
            } else {
                printf(DLLNAME ": SUCCESS: Triggered and caught %d exceptions.\n", exceptionsTriggered);
            }
        } else {
            printf(DLLNAME ": ERROR: Mismatch. %d exceptions triggered, %d exceptions handled.\n", exceptionsTriggered, exceptionsHandled);
        }
        fflush(stdout);
    }
    return TRUE;
}

//
///////////////////////////////////////////////////////////////// End of File.
